################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
import numpy as np
import pytest
import sympy
from scipy.linalg import fractional_matrix_power

from orquestra.quantum import circuits
from orquestra.quantum.circuits import XX, YY, ZZ, Circuit
from orquestra.quantum.evolution import (
    _generate_circuit_sequence,
    time_evolution,
    time_evolution_derivatives,
    time_evolution_for_term,
)
from orquestra.quantum.operators import PauliSum, PauliTerm
from orquestra.quantum.utils import compare_unitary

TUPLE_TERM_TO_ORQUESTRA_GATE = {
    ((0, "X"), (1, "X")): XX,
    ((0, "Y"), (1, "Y")): YY,
    ((0, "Z"), (1, "Z")): ZZ,
}


def _orquestra_exponentiate_qubit_hamiltonian_term(term, time, trotter_order):
    base_exponent = 2 * time / trotter_order / np.pi

    base_gate = TUPLE_TERM_TO_ORQUESTRA_GATE[tuple(sorted(term.operations))]
    # This introduces a phase to the gate, but that's fine
    # since `compare_unitary` accounts for that
    mat = base_gate(np.pi)(0, 1).lifted_matrix(2)
    if term.coefficient.real == term.coefficient:
        coefficient = term.coefficient.real
    else:
        raise ValueError("Cannot raise matrix to the power of a complex coefficinet")
    zquant_mat = fractional_matrix_power(mat, coefficient * base_exponent)

    return zquant_mat


def _orquestra_exponentiate_hamiltonian(hamiltonian, time, trotter_order):
    ops = []
    for term in hamiltonian.terms:
        mat = _orquestra_exponentiate_qubit_hamiltonian_term(term, time, trotter_order)
        ops.append(
            circuits.CustomGateDefinition(
                gate_name="custom_a",
                matrix=sympy.Matrix(mat),
                params_ordering=(),
            )()(0, 1)
        )

    circuit = Circuit(operations=ops * trotter_order)

    return circuit


@pytest.mark.parametrize(
    "term, time, expected_unitary",
    [
        (PauliTerm("X0*X1"), np.pi, -np.eye(4)),
        (
            PauliTerm("0.5*Y0*Y1"),
            np.pi,
            np.diag([1j, -1j, -1j, 1j])[::-1],
        ),
        (PauliTerm("Z0*Z1"), np.pi, -np.eye(4)),
    ],
)
class TestTimeEvolutionOfTerm:
    def test_evolving_pauli_term_with_numerical_time_gives_correct_unitary(
        self, term, time, expected_unitary
    ):
        actual_unitary = time_evolution_for_term(term, time).to_unitary()
        np.testing.assert_array_almost_equal(actual_unitary, expected_unitary)

    def test_evolving_pauli_term_with_symbolic_time_gives_correct_unitary(
        self, term, time, expected_unitary
    ):
        time_symbol = sympy.Symbol("t")
        symbol_map = {time_symbol: time}
        evolution_circuit = time_evolution_for_term(term, time_symbol)

        actual_unitary = evolution_circuit.bind(symbol_map).to_unitary()
        np.testing.assert_array_almost_equal(actual_unitary, expected_unitary)


def test_complex_coefficent_throws_error():
    with pytest.raises(ValueError):
        time_evolution_for_term(PauliTerm("X0", 1j), 1)


class TestTimeEvolutionOfConstantTerm:
    # This test is added to make sure that constant terms in qubit operators
    # do not cause errors.
    def test_evolving_constant_term_qubit_operator_gives_empty_circuit(self):
        evolution_circuit = time_evolution_for_term(PauliTerm.identity() * 3, np.pi)
        assert evolution_circuit == circuits.Circuit()


class TestTimeEvolutionOfPauliSum:
    @pytest.fixture
    def hamiltonian(self):
        return PauliSum("X0*X1 + 0.5*Y0*Y1 + 0.3*Z0*Z1")

    @pytest.mark.parametrize("time", [0.1, 0.4, 1.0])
    @pytest.mark.parametrize("order", [1, 2, 3])
    def test_evolution_with_numerical_time_produces_correct_result(
        self, hamiltonian, time, order
    ):
        expected_orquestra_circuit = _orquestra_exponentiate_hamiltonian(
            hamiltonian, time, order
        )

        reference_unitary = expected_orquestra_circuit.to_unitary()
        unitary = time_evolution(hamiltonian, time, trotter_order=order).to_unitary()

        assert compare_unitary(unitary, reference_unitary, tol=1e-10)

    @pytest.mark.parametrize("time_value", [0.1, 0.4, 1.0])
    @pytest.mark.parametrize("order", [1, 2, 3])
    def test_time_evolution_with_symbolic_time_produces_correct_unitary(
        self, hamiltonian, time_value, order
    ):
        time_symbol = sympy.Symbol("t")
        symbols_map = {time_symbol: time_value}

        expected_orquestra_circuit = _orquestra_exponentiate_hamiltonian(
            hamiltonian, time_value, order
        )

        reference_unitary = expected_orquestra_circuit.to_unitary()

        unitary = (
            time_evolution(hamiltonian, time_symbol, trotter_order=order)
            .bind(symbols_map)
            .to_unitary()
        )

        assert compare_unitary(unitary, reference_unitary, tol=1e-10)


class TestGeneratingCircuitSequence:
    @pytest.mark.parametrize(
        "repeated_circuit, different_circuit, length, position, expected_result",
        [
            (
                circuits.Circuit([circuits.X(0), circuits.Y(1)]),
                circuits.Circuit([circuits.Z(1)]),
                5,
                1,
                circuits.Circuit(
                    [
                        *[circuits.X(0), circuits.Y(1)],
                        circuits.Z(1),
                        *([circuits.X(0), circuits.Y(1)] * 3),
                    ]
                ),
            ),
            (
                circuits.Circuit([circuits.RX(0.5)(1)]),
                circuits.Circuit([circuits.CNOT(0, 2)]),
                3,
                0,
                circuits.Circuit(
                    [circuits.CNOT(0, 2), circuits.RX(0.5)(1), circuits.RX(0.5)(1)]
                ),
            ),
            (
                circuits.Circuit([circuits.RX(0.5)(1)]),
                circuits.Circuit([circuits.CNOT(0, 2)]),
                3,
                2,
                circuits.Circuit(
                    [
                        circuits.RX(0.5)(1),
                        circuits.RX(0.5)(1),
                        circuits.CNOT(0, 2),
                    ]
                ),
            ),
        ],
    )
    def test_produces_correct_sequence(
        self, repeated_circuit, different_circuit, position, length, expected_result
    ):
        actual_result = _generate_circuit_sequence(
            repeated_circuit, different_circuit, length, position
        )

        assert actual_result == expected_result

    @pytest.mark.parametrize("length, invalid_position", [(5, 5), (4, 6)])
    def test_raises_error_if_position_is_larger_or_equal_to_length(
        self, length, invalid_position
    ):
        repeated_circuit = circuits.Circuit([circuits.X(0)])
        different_circuit = circuits.Circuit([circuits.Y(0)])

        with pytest.raises(ValueError):
            _generate_circuit_sequence(
                repeated_circuit, different_circuit, length, invalid_position
            )


class TestTimeEvolutionDerivatives:
    @pytest.fixture
    def hamiltonian(self):
        return PauliSum("X0*X1 + 0.5*Y0*Y1 + 0.3*Z0*Z1")

    @pytest.mark.parametrize("time", [0.4, sympy.Symbol("t")])
    def test_gives_correct_number_of_derivatives_and_factors(self, time, hamiltonian):

        order = 3
        reference_factors_1 = [1.0 / order, 0.5 / order, 0.3 / order] * 3
        reference_factors_2 = [-1.0 * x for x in reference_factors_1]

        derivatives, factors = time_evolution_derivatives(
            hamiltonian, time, trotter_order=order
        )

        assert len(derivatives) == order * 2 * len(hamiltonian.terms)
        assert len(factors) == order * 2 * len(hamiltonian.terms)
        assert factors[0:18:2] == reference_factors_1
        assert factors[1:18:2] == reference_factors_2



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
"""Test cases for sympy_expressions module."""
import pytest
import sympy

from orquestra.quantum.circuits.symbolic.expressions import FunctionCall, Symbol
from orquestra.quantum.circuits.symbolic.sympy_expressions import (
    expression_from_sympy,
    is_addition_of_negation,
    is_multiplication_by_reciprocal,
)


@pytest.mark.parametrize("number", [3, 4.0, 1j, 3.0 - 2j])
def test_native_numbers_are_preserved(number):
    assert expression_from_sympy(number) == number


@pytest.mark.parametrize(
    "sympy_symbol, expected_symbol",
    [
        (sympy.Symbol("theta"), Symbol("theta")),
        (sympy.Symbol("x"), Symbol("x")),
        (sympy.Symbol("c_i"), Symbol("c_i")),
    ],
)
def test_symbols_are_converted_to_instance_of_symbol_class(
    sympy_symbol, expected_symbol
):
    assert expression_from_sympy(sympy_symbol) == expected_symbol


@pytest.mark.parametrize(
    "sympy_number, expected_number, expected_class",
    [
        (sympy.sympify(2), 2, int),
        (sympy.sympify(-2.5), -2.5, float),
        (sympy.Rational(3, 8), 0.375, float),
    ],
)
def test_sympy_numbers_are_converted_to_corresponding_native_number(
    sympy_number, expected_number, expected_class
):
    native_number = expression_from_sympy(sympy_number)
    assert native_number == expected_number
    assert isinstance(native_number, expected_class)


def test_imaginary_unit_is_converted_to_1j():
    assert expression_from_sympy(sympy.I) == 1j


# In below functions we explicitly construct Add and Mul objects
# because arithmetic operations on sympy expressions may perform
# additional evaluation which may circumvent our expectations.
@pytest.mark.parametrize(
    "sympy_add, expected_args",
    [
        (sympy.Add(1, 2, 3, evaluate=False), (1, 2, 3)),
        (sympy.Add(sympy.Symbol("x"), 1, evaluate=False), (Symbol("x"), 1)),
        (
            sympy.Add(
                sympy.Symbol("x"),
                sympy.Symbol("y"),
                sympy.Symbol("z"),
                evaluate=False,
            ),
            (Symbol("x"), Symbol("y"), Symbol("z")),
        ),
    ],
)
def test_sympy_add_is_converted_to_function_call_with_add_operation(
    sympy_add, expected_args
):
    assert expression_from_sympy(sympy_add) == FunctionCall("add", expected_args)


@pytest.mark.parametrize(
    "sympy_mul, expected_args",
    [
        (sympy.Mul(4, 2, 3, evaluate=False), (4, 2, 3)),
        (sympy.Mul(sympy.Symbol("x"), 2, evaluate=False), (Symbol("x"), 2)),
        (
            sympy.Mul(
                sympy.Symbol("x"),
                sympy.Symbol("y"),
                sympy.Symbol("z"),
                evaluate=False,
            ),
            (Symbol("x"), Symbol("y"), Symbol("z")),
        ),
    ],
)
def test_sympy_mul_is_converted_to_function_call_with_mul_operation(
    sympy_mul, expected_args
):
    assert expression_from_sympy(sympy_mul) == FunctionCall("mul", expected_args)


@pytest.mark.parametrize(
    "sympy_multiplication",
    [
        sympy.Symbol("x") / sympy.Symbol("y"),
        sympy.Symbol("x") / (sympy.Symbol("z") + 1),
    ],
)
def test_mul_from_division_is_classified_as_multiplication_by_reciprocal(
    sympy_multiplication,
):
    assert is_multiplication_by_reciprocal(sympy_multiplication)


@pytest.mark.parametrize(
    "sympy_multiplication",
    [
        sympy.Symbol("x") * sympy.Symbol("y"),
        2 * sympy.Symbol("theta"),
        sympy.Symbol("x") * sympy.Symbol("y") * sympy.Symbol("z"),
    ],
)
def test_mul_not_from_division_is_not_classified_as_multiplication_by_reciprocal(
    sympy_multiplication,
):
    # Note: obviously you can manually construct multiplication that would
    # be classified as multiplication by reciprocal. The bottom line of this
    # test is: usual, simple multiplications are multiplications, not divisions.
    assert not is_multiplication_by_reciprocal(sympy_multiplication)


@pytest.mark.parametrize(
    "sympy_multiplication, expected_args",
    [
        (sympy.Symbol("x") / sympy.Symbol("y"), (Symbol("x"), Symbol("y"))),
        (
            sympy.Symbol("x") / (sympy.Add(sympy.Symbol("z"), 1, evaluate=False)),
            (Symbol("x"), FunctionCall("add", (Symbol("z"), 1))),
        ),
    ],
)
def test_division_is_converted_into_div_fn_call_instead_of_multiplication_by_reciprocal(
    sympy_multiplication, expected_args
):
    # Important note about sympy: there is no Div operator (as opposed to
    # e.g. Mul). The division on sympy expressions actually produces Mul
    # objects, in which second operand is a reciprocal of the original one.
    # We need to deal with this case, otherwise converting anything that
    # contains division will result in very confusing expressions.
    assert expression_from_sympy(sympy_multiplication) == FunctionCall(
        "div", expected_args
    )


@pytest.mark.parametrize(
    "sympy_addition",
    [
        sympy.Symbol("x") - sympy.Symbol("y"),
        sympy.Symbol("x") - 1 / sympy.Symbol("y"),
        1 - sympy.Symbol("x")
        # Note: negation of previous case would fail, since in that case
        # sympy would make Add(-1, Symbol("x")) out of it, unlike in other
        # cases where it produces e.g.
        # Add(Symbol("x", Mul(Symbol("y"), -1))).
    ],
)
def test_add_resulting_from_subtraction_is_classified_as_addition_of_negation(
    sympy_addition,
):
    assert is_addition_of_negation(sympy_addition)


@pytest.mark.parametrize(
    "sympy_addition",
    [sympy.Symbol("x") + sympy.Symbol("y"), sympy.Symbol("x") + 10],
)
def test_add_not_resulting_from_subtraction_is_not_classified_as_addition_of_negation(
    sympy_addition,
):
    assert not is_addition_of_negation(sympy_addition)


@pytest.mark.parametrize(
    "sympy_addition, expected_args",
    [
        (sympy.Symbol("x") - sympy.Symbol("y"), (Symbol("x"), Symbol("y"))),
        (1 - sympy.Symbol("x"), (1, Symbol("x"))),
        (
            1 - sympy.Symbol("x") * sympy.Symbol("y"),
            (1, FunctionCall("mul", (Symbol("x"), Symbol("y")))),
        ),
    ],
)
def test_add_resulting_from_subtraction_is_converted_to_sub_function_call(
    sympy_addition, expected_args
):
    assert expression_from_sympy(sympy_addition) == FunctionCall("sub", expected_args)


@pytest.mark.parametrize(
    "sympy_power, expected_args",
    [
        (sympy.Pow(sympy.Symbol("x"), 2), (Symbol("x"), 2)),
        (sympy.Pow(2, sympy.Symbol("x")), (2, Symbol("x"))),
        (
            sympy.Pow(sympy.Symbol("x"), sympy.Symbol("y")),
            (Symbol("x"), Symbol("y")),
        ),
    ],
)
def test_sympy_pow_is_converted_to_pow_function_call(sympy_power, expected_args):
    assert expression_from_sympy(sympy_power) == FunctionCall("pow", expected_args)


@pytest.mark.parametrize(
    "sympy_power, expected_denominator",
    [
        (sympy.Pow(sympy.Symbol("x"), -1), Symbol("x")),
        (
            sympy.Pow(
                sympy.Add(sympy.Symbol("x"), sympy.Symbol("y"), evaluate=False), -1
            ),
            FunctionCall("add", (Symbol("x"), Symbol("y"))),
        ),
    ],
)
def test_sympy_power_with_negative_one_exponent_gets_converted_to_division(
    sympy_power, expected_denominator
):
    assert expression_from_sympy(sympy_power) == FunctionCall(
        "div", (1, expected_denominator)
    )


@pytest.mark.parametrize(
    "sympy_function_call, expected_function_call",
    [
        (sympy.cos(2), FunctionCall("cos", (2,))),
        (sympy.sin(sympy.Symbol("theta")), FunctionCall("sin", (Symbol("theta"),))),
        (sympy.exp(sympy.Symbol("x")), FunctionCall("exp", (Symbol("x"),))),
    ],
)
def test_sympy_fn_calls_are_converted_to_fn_call_object_with_appropriate_fn_name(
    sympy_function_call, expected_function_call
):
    assert expression_from_sympy(sympy_function_call) == expected_function_call



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
from dataclasses import dataclass
from functools import singledispatch
from numbers import Complex
from typing import Iterable, Tuple

import numpy as np
import sympy

from ..typing import ParameterizedVector
from ._operations import Parameter, get_free_symbols, sub_symbols


@singledispatch
def _is_real(number: Complex):
    return number.imag == 0


@_is_real.register
def _is_sympy_number_real(number: sympy.Number):
    return number.is_real


@dataclass(frozen=True)
class MultiPhaseOperation:
    """Operation applying distinct phase to each wavefunction component.

    MultiPhaseOperation with parameters theta_1, theta_2, .... theta_2^N,
    transforms a N qubit wavefunction (psi_1, psi_2, ..., psi_2^N)
    into (exp(i theta_1)psi_1, exp(i theta_2) psi_2, ..., exp(i theta_2^N) psi_2^N).
    """

    params: Tuple[Parameter, ...]

    def __post_init__(self):
        if any(
            isinstance(param, Complex) and not _is_real(param) for param in self.params
        ):
            raise ValueError("MultiPhaseOperation supports only real parameters.")

    @property
    def qubit_indices(self) -> Tuple[int, ...]:
        n_qubits = int(np.log2(len(self.params)))
        return tuple(range(n_qubits))

    def bind(self, symbols_map) -> "MultiPhaseOperation":
        return self.replace_params(
            tuple(sub_symbols(param, symbols_map) for param in self.params)
        )

    def replace_params(
        self, new_params: Tuple[Parameter, ...]
    ) -> "MultiPhaseOperation":
        return MultiPhaseOperation(new_params)

    def apply(self, amplitude_vector: ParameterizedVector) -> ParameterizedVector:
        if len(amplitude_vector) != len(self.params):
            raise ValueError(
                f"MultiPhaseOperation with {len(self.params)} params cannot be "
                f"applied to amplitude_vector of length {len(amplitude_vector)}."
            )

        try:
            exp_params = np.exp(np.asarray(self.params, dtype=float) * 1j)
        except TypeError as e:
            raise RuntimeError(
                "MultiPhaseOperation can only be applied only if all symbolic "
                "parameters are bound to real numbers."
            ) from e
        return np.multiply(np.asarray(amplitude_vector), exp_params)

    @property
    def free_symbols(self) -> Iterable[sympy.Symbol]:
        """Unbound symbols in the gate matrix.

        Examples:

        - an `H` gate has no free symbols
        - a `RX(np.pi)` gate has no free symbols
        - a `RX(sympy.Symbol("theta"))` gate has a single free symbol `theta`
        - a `RX(sympy.sympify("theta * alpha"))` gate has two free symbols, `alpha` and
            `theta`
        - a `RX(sympy.sympify("theta * alpha")).bind({sympy.Symbol("theta"): 0.42})`
            gate has one free symbol, `alpha`
        """
        return get_free_symbols(self.params)



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
from __future__ import annotations

import copy
import json
from collections import Counter
from typing import Dict, Iterable, List, Optional, Sequence, TextIO, Tuple

import numpy as np

from ..distributions import MeasurementOutcomeDistribution
from ..operators import PauliRepresentation
from ..typing import AnyPath
from ..utils import convert_tuples_to_bitstrings, sample_from_probability_distribution
from .expectation_values import ExpectationValues
from .parities import check_parity_of_vector


def convert_bitstring_to_int(bitstring: Sequence[int]) -> int:
    """Convert a bitstring to an integer.

    Args:
        bitstring (list): A list of integers.
    Returns:
        int: The value of the bitstring, where the first bit in the least
            significant (little endian).
    """
    return int("".join(str(bit) for bit in bitstring[::-1]), 2)


def _convert_bitstrings_to_vector(bitstrings: Iterable[str]) -> np.ndarray:
    """Converts bitstrings to vector so parity can be checked."""
    n_qubits = len([*bitstrings][0])
    all_bits = "".join(bitstrings)
    bitstring_1d_array = np.frombuffer(all_bits.encode("utf-8"), "u1") - ord("0")
    return bitstring_1d_array.astype(int).reshape(-1, n_qubits)


def get_expectation_value_from_frequencies(
    marked_qubits: Iterable[int], bitstring_frequencies: Dict[str, int]
) -> float:
    """Get the expectation value the product of Z operators on selected qubits
    from bitstring frequencies.

    Args:
        marked_qubits: The qubits that the Z operators act on.
        bitstring_frequences: The frequencies of the bitstrings.

    Returns:
        The expectation value of the product of Z operators on selected qubits.
    """

    parity = (
        check_parity_of_vector(
            _convert_bitstrings_to_vector(bitstring_frequencies.keys()),
            marked_qubits,
        )
        * 2
        - 1
    )
    num_measurements = sum(bitstring_frequencies.values())
    expectation_values: np.ndarray = (
        np.fromiter(bitstring_frequencies.values(), dtype=int)
        * parity
        / num_measurements
    )

    # The item method converts a numpy float to a native Python float
    return expectation_values.sum().item()


def _check_sample_elimination(
    samples: Counter,
    bitstring_samples: List[Tuple[int, ...]],
    leftover_distribution: MeasurementOutcomeDistribution,
) -> Counter:
    """This is a function that checks that all elements in samples
    are present in bitstring_samples. If they are not, we eliminate the
    elements not in bitstring samples, set their probability to 0 in
    leftover_distribution and resample the appropriate number of times.
    Then, we re-check the new samples.
    Args:
        samples: The bitstrings to eliminate and how many times to eliminate them
        bitstring_samples: the bitstring distribution from where the bitstrings
                           should be removed
        leftover_distribution: the distribution from which samples to eliminate are
                               sampled
    Returns:
        correct_samples: A sample object that only contains bitstrings that can actually
                     be removed from bitstring_samples
    """
    bitstring_counts = Counter(bitstring_samples)

    nresamples = 1  # Initializing so that the loop starts
    corrected_leftover_distribution = MeasurementOutcomeDistribution(
        leftover_distribution.distribution_dict
    )
    correct_samples = samples.copy()
    while nresamples != 0:
        new_samples = None
        nresamples = 0
        for sample in correct_samples:
            bitstring = tuple([int(measurement_value) for measurement_value in sample])
            if correct_samples[sample] > bitstring_counts[bitstring]:
                nresamples = correct_samples[sample] - bitstring_counts[bitstring]
                correct_samples[sample] = bitstring_counts[bitstring]
                distribution_dict = corrected_leftover_distribution.distribution_dict
                distribution_dict[sample] = 0
                corrected_leftover_distribution = MeasurementOutcomeDistribution(
                    distribution_dict, True
                )
                new_samples = sample_from_probability_distribution(
                    corrected_leftover_distribution.distribution_dict, nresamples
                )
                correct_samples = correct_samples + new_samples
                break

    return correct_samples


class Measurements:
    """A class representing measurements from a quantum circuit. The bitstrings variable
    represents the internal data structure of the Measurements class. It is expressed as
    a list of tuples wherein each tuple is a measurement and the value of the tuple at a
    given index is the measured bit-value of the qubit (indexed from 0 -> N-1)"""

    def __init__(self, bitstrings: Optional[List[Tuple[int, ...]]] = None):
        if bitstrings is None:
            self.bitstrings = []
        else:
            self.bitstrings = bitstrings

    @classmethod
    def from_counts(cls, counts: Dict[str, int]):
        """Create an instance of the Measurements class from a dictionary

        Args:
            counts: mapping of bitstrings to integers representing the number of times
                the bitstring was measured
        """
        measurements = cls()
        measurements.add_counts(counts)
        return measurements

    @classmethod
    def get_measurements_representing_distribution(
        cls,
        measurement_outcome_distribution: MeasurementOutcomeDistribution,
        number_of_samples: int,
    ):
        """Create an instance of the Measurements class that exactly (or as closely as
        possible) resembles the input bitstring distribution.

        Args:
            measurement_outcome_distribution: the bitstring distribution to be sampled
            number_of_samples: the number of measurements
        """
        distribution = copy.deepcopy(measurement_outcome_distribution.distribution_dict)

        bitstring_samples = []
        # Rounding gives the closest integer to the observed frequency
        for state in distribution:
            bitstring = tuple([int(measurement_value) for measurement_value in state])

            bitstring_samples += [bitstring] * int(
                round(distribution[state] * number_of_samples)
            )

        # If frequencies are inconsistent with number of samples, we may need to
        # add or delete samples. The bitstrings to correct are chosen at random,
        # giving more weight to those with non-integer part closest to 0.5
        if len(bitstring_samples) != number_of_samples:
            leftover_distribution = MeasurementOutcomeDistribution(
                {
                    states: 0.5
                    - abs(0.5 - (distribution[states] * number_of_samples) % 1)
                    for states in distribution.keys()
                },
                True,
            )

            samples = sample_from_probability_distribution(
                leftover_distribution.distribution_dict,
                abs(number_of_samples - len(bitstring_samples)),
            )
            if number_of_samples - len(bitstring_samples) > 0:
                for sample in samples:
                    bitstring_samples += [
                        tuple([int(measurement_value) for measurement_value in sample])
                    ] * samples[sample]
            else:
                # Eliminating samples: need to ensure they are present in the
                # bitstring_samples list
                samples = _check_sample_elimination(
                    samples, bitstring_samples, leftover_distribution
                )
                for sample in samples:
                    for _ in range(samples[sample]):
                        bitstring_samples.remove(
                            tuple(
                                [int(measurement_value) for measurement_value in sample]
                            )
                        )

        return cls(bitstring_samples)

    @classmethod
    def load_from_file(cls, file: TextIO):
        """Load a set of measurements from file

        Args:
            file (str or file-like object): the name of the file, or a file-like object
        """
        if isinstance(file, str):
            with open(file, "r") as f:
                data = json.load(f)
        else:
            data = json.load(file)

        bitstrings = []
        for bitstring in data["bitstrings"]:
            bitstrings.append(tuple(bitstring))

        return cls(bitstrings=bitstrings)

    def save(self, filename: AnyPath):
        """Serialize the Measurements object into a file in JSON format.

        Args:
            filename (string): filename to save the data to
        """
        data = {
            "counts": self.get_counts(),
            # This step is required if bistrings contain np.int8 instead of regular int.
            "bitstrings": [
                list(map(int, list(bitstring))) for bitstring in self.bitstrings
            ],
        }
        with open(filename, "w") as f:
            f.write(json.dumps(data, indent=2))

    def get_counts(self):
        """Get the measurements as a histogram

        Returns:
            A dictionary mapping bitstrings to integers representing the number of times
            the bitstring was measured
        """
        bitstrings = convert_tuples_to_bitstrings(self.bitstrings)
        return dict(Counter(bitstrings))

    def add_counts(self, counts: Dict[str, int]):
        """Add measurements from a histogram

        Args:
            counts: mapping of bitstrings to integers representing the number of times
                the bitstring was measured
                NOTE: bitstrings are also indexed from 0 -> N-1, where the "001"
                bitstring represents a measurement of qubit 2 in the 1 state
        """
        for bitstring in counts.keys():
            measurement = []
            for bitvalue in bitstring:
                measurement.append(int(bitvalue))

            self.bitstrings += [tuple(measurement)] * counts[bitstring]

    def get_distribution(self) -> MeasurementOutcomeDistribution:
        """Get the normalized probability distribution representing the measurements

        Returns:
            distribution: bitstring distribution based on the frequency of measurements
        """
        counts = self.get_counts()
        num_measurements = len(self.bitstrings)

        distribution = {}
        for bitstring in counts.keys():
            distribution[bitstring] = counts[bitstring] / num_measurements

        return MeasurementOutcomeDistribution(distribution)

    def get_expectation_values(
        self, ising_operator: PauliRepresentation, use_bessel_correction: bool = False
    ) -> ExpectationValues:
        """Get the expectation values of an operator from the measurements.

        Args:
            ising_operator: the operator
            use_bessel_correction: Whether to use Bessel's correction when
                when estimating the covariance of operators. Using the
                correction provides an unbiased estimate for covariances, but
                diverges when only one sample is taken.

        Returns:
            expectation values of each term in the operator
        """
        # We require operator to be ising because measurements are always performed in
        # the Z basis, so we need the operator to be Ising (containing only Z terms).
        # A general Qubit Operator could have X or Y terms which don’t get directly
        # measured.
        if not ising_operator.is_ising:
            raise TypeError("Input operator is not ising.")

        # Count number of occurrences of bitstrings
        bitstring_frequencies = self.get_counts()
        num_measurements = len(self.bitstrings)

        # Perform weighted average
        expectation_values_list = [
            term.coefficient
            * get_expectation_value_from_frequencies(term.qubits, bitstring_frequencies)
            for term in ising_operator.terms
        ]
        expectation_values = np.array(expectation_values_list)

        correlations = np.zeros((len(ising_operator.terms),) * 2, dtype=complex)
        for i, first_term in enumerate(ising_operator.terms):
            correlations[i, i] = first_term.coefficient**2
            for j in range(i):
                second_term = ising_operator.terms[j]
                marked_qubits = first_term.qubits.symmetric_difference(
                    second_term.qubits
                )
                correlations[i, j] = (
                    first_term.coefficient
                    * second_term.coefficient
                    * get_expectation_value_from_frequencies(
                        marked_qubits, bitstring_frequencies
                    )
                )
                correlations[j, i] = correlations[i, j]

        denominator = (
            num_measurements - 1 if use_bessel_correction else num_measurements
        )

        estimator_covariances = (
            correlations
            - expectation_values[:, np.newaxis] * expectation_values[np.newaxis, :]
        ) / denominator

        return ExpectationValues(
            expectation_values, [correlations], [estimator_covariances]
        )



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
import random
import unittest

import numpy as np

from orquestra.quantum.operators import PauliSum, PauliTerm, get_sparse_operator
from orquestra.quantum.utils import RNDSEED
from orquestra.quantum.wavefunction import Wavefunction


class Testpaulioperator(unittest.TestCase):
    def test_build_paulioperator_from_coeffs_and_labels(self):
        from orquestra.quantum.operators._utils import (
            get_pauliop_from_coeffs_and_labels,
        )

        # Given
        test_op = PauliTerm("(3j)*Y0*X1*Z2*X4")
        coeffs = [3.0j]
        labels = [[2, 1, 3, 0, 1]]

        # When
        build_op = get_pauliop_from_coeffs_and_labels(coeffs, labels)

        # Then
        self.assertEqual(test_op, build_op)

    def test_pauliop_matrix_converion(self):
        from orquestra.quantum.operators._utils import get_pauliop_from_matrix

        # Given
        m = 4
        n = 2**m
        TOL = 10**-15
        random.seed(RNDSEED)
        A = np.array([[random.uniform(-1, 1) for x in range(n)] for y in range(n)])

        # When
        A_pauliop = get_pauliop_from_matrix(A)
        A_pauliop_matrix = np.array(get_sparse_operator(A_pauliop).todense())
        test_matrix = A_pauliop_matrix - A

        # Then
        for row in test_matrix:
            for elem in row:
                self.assertEqual(abs(elem) < TOL, True)

    def test_generate_random_pauliop(self):
        from orquestra.quantum.operators._utils import generate_random_pauliop

        # Given
        nqubits = 4
        nterms = 5
        nlocality = 2
        max_coeff = 1.5
        fixed_coeff = False

        # When
        op = generate_random_pauliop(nqubits, nterms, nlocality, max_coeff, fixed_coeff)
        # Then
        self.assertEqual(len(op.terms), nterms)
        for term in op.terms:
            self.assertLess(sorted(term.qubits)[-1], nqubits)  # max qubit < nqubits
            self.assertEqual(len(term), nlocality)
            self.assertLessEqual(np.abs(term.coefficient), max_coeff)

        # Given
        fixed_coeff = True
        # When
        op = generate_random_pauliop(nqubits, nterms, nlocality, max_coeff, fixed_coeff)
        # Then
        self.assertEqual(len(op.terms), nterms)
        for term in op.terms:
            self.assertEqual(np.abs(term.coefficient), max_coeff)

    def test_evaluate_operator(self):
        from orquestra.quantum.measurements import ExpectationValues
        from orquestra.quantum.operators._utils import evaluate_operator

        # Given
        qubit_op = PauliSum("0.5 + 0.5*Z1")
        expectation_values = ExpectationValues([0.5, 0.5])
        # When
        value_estimate = evaluate_operator(qubit_op, expectation_values)
        # Then
        self.assertAlmostEqual(value_estimate, 0.5)

    def test_evaluate_operator_list(self):
        from orquestra.quantum.measurements import ExpectationValues
        from orquestra.quantum.operators._utils import evaluate_operator_list

        # Given
        qubit_op_list = [PauliSum("0.5 + 0.5*Z1"), PauliSum("0.3*X1 + 0.2*Y2")]
        expectation_values = ExpectationValues([0.5, 0.5, 0.4, 0.6])
        # When
        value_estimate = evaluate_operator_list(qubit_op_list, expectation_values)
        # Then
        self.assertAlmostEqual(value_estimate, 0.74)

    def test_reverse_qubit_order(self):
        from orquestra.quantum.operators._utils import reverse_qubit_order

        # Given
        op1 = PauliTerm("Z0*Z1")
        op2 = PauliTerm("Z1*Z0")

        # When/Then
        self.assertEqual(op1, reverse_qubit_order(op2))

        # Given
        op1 = PauliTerm("Z0")
        op2 = PauliTerm("Z1")

        # When/Then
        self.assertEqual(op1, reverse_qubit_order(op2, n_qubits=2))
        self.assertEqual(op2, reverse_qubit_order(op1, n_qubits=2))

    def test_get_expectation_value(self):
        from orquestra.quantum.operators._utils import get_expectation_value

        """Check <Z0> and <Z1> for the state |100>"""
        # Given
        wf = Wavefunction([0, 1, 0, 0, 0, 0, 0, 0])
        op1 = PauliTerm("Z0")
        op2 = PauliTerm("Z1")

        # When
        exp_op1 = get_expectation_value(op1, wf, True)
        exp_op2 = get_expectation_value(op2, wf, True)

        # Then
        self.assertAlmostEqual(-1, exp_op1)
        self.assertAlmostEqual(1, exp_op2)



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
"""Test cases for _builtin_gates_module."""
import pytest

from orquestra.quantum.circuits import _builtin_gates


class TestBuiltinGatesProperties:
    @pytest.mark.parametrize(
        "gate",
        [
            _builtin_gates.X,
            _builtin_gates.Y,
            _builtin_gates.Z,
            _builtin_gates.CZ,
            _builtin_gates.CNOT,
            _builtin_gates.I,
            _builtin_gates.T,
            _builtin_gates.H,
            _builtin_gates.S,
            _builtin_gates.SWAP,
            _builtin_gates.ISWAP,
            _builtin_gates.RX(0.5),
            _builtin_gates.RY(1),
            _builtin_gates.RZ(0.5),
            _builtin_gates.RH(1.5),
            _builtin_gates.U3(0.5, -3.14, 0),
            _builtin_gates.XX(0.1),
            _builtin_gates.YY(0.2),
            _builtin_gates.ZZ(0.3),
            _builtin_gates.PHASE(1),
            _builtin_gates.CPHASE(0.1),
            _builtin_gates.GPi(0.1),
            _builtin_gates.GPi2(0.1),
            _builtin_gates.MS(0.1, 0.2),
            _builtin_gates.Delay(0.5),
        ],
    )
    def test_gates_matrix_equals_its_adjoint_iff_gate_is_hermitian(self, gate):
        assert (gate.matrix == gate.matrix.adjoint()) == gate.is_hermitian



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
from json import load
from os import remove

import numpy as np
import pytest

from orquestra.quantum.circuits import CNOT, Circuit, H, X
from orquestra.quantum.distributions import MeasurementOutcomeDistribution
from orquestra.quantum.runners import SymbolicSimulator
from orquestra.quantum.runners.trackers import MeasurementTrackingBackend


@pytest.fixture(scope="function")
def backend():
    backend = MeasurementTrackingBackend(SymbolicSimulator(), "tracker_test")
    return backend


class TestMeasurementTrackingBackend:
    """First we repeat tests from the QuantumBackendTests class
    modified to work with MeasurementTrackingBackend.

    Then we check that serialization performs as expected.
    """

    def test_run_circuit_and_measure_correct_indexing(self, backend):
        # Note: this test may fail with noisy devices
        # Given
        circuit = Circuit([X(0), X(0), X(1), X(1), X(2)])
        n_samples = 100

        try:
            # When
            measurements = backend.run_and_measure(circuit, n_samples)

            # Then (since SPAM error could result in unexpected bitstrings, we make
            # sure the most common bitstring is the one we expect)
            counts = measurements.get_counts()
            assert max(counts, key=counts.get) == "001"
            assert backend.inner_backend.n_circuits_executed == 1
            assert backend.inner_backend.n_jobs_executed == 1
        finally:
            # Cleanup
            remove(backend.raw_data_file_name)

    @pytest.mark.parametrize("n_samples", [-1, 0])
    def test_run_circuit_and_measure_fails_for_invalid_n_samples(
        self, backend, n_samples
    ):
        # Given
        circuit = Circuit([X(0), X(0), X(1), X(1), X(2)])

        # When
        with pytest.raises(ValueError):
            backend.run_and_measure(circuit, n_samples)
            # Cleanup
            remove(backend.raw_data_file_name)

    def test_run_batch_and_measure(self, backend):
        # Given
        circuit = Circuit([X(0), X(0), X(1), X(1), X(2)])
        n_samples = 100
        number_of_circuits = 25

        # When
        n_samples_per_circuit = [n_samples] * number_of_circuits
        try:
            measurements_set = backend.run_batch_and_measure(
                [circuit] * number_of_circuits, n_samples_per_circuit
            )

            for measurements in measurements_set:
                counts = measurements.get_counts()
                assert max(counts, key=counts.get) == "001"
            assert backend.inner_backend.n_circuits_executed == number_of_circuits
            assert backend.inner_backend.n_jobs_executed <= number_of_circuits
        finally:
            # Cleanup
            remove(backend.raw_data_file_name)

    def test_run_batch_and_measure_n_samples(self, backend):
        # Note: this test may fail with noisy devices
        # Given
        first_circuit = Circuit([X(0), X(0), X(1), X(1), X(2)])
        second_circuit = Circuit([X(0), X(1), X(2)])
        n_samples_per_circuit = [100, 105]

        try:
            # When
            measurements_set = backend.run_batch_and_measure(
                [first_circuit, second_circuit], n_samples_per_circuit
            )

            # Then (since SPAM error could result in unexpected bitstrings, we make
            # sure the most common bitstring is the one we expect)
            counts = measurements_set[0].get_counts()
            assert max(counts, key=counts.get) == "001"
            counts = measurements_set[1].get_counts()
            assert max(counts, key=counts.get) == "111"

            assert len(measurements_set[0].bitstrings) == n_samples_per_circuit[0]
            assert len(measurements_set[1].bitstrings) == n_samples_per_circuit[1]

            assert backend.inner_backend.n_circuits_executed == 2
        finally:
            # Cleanup
            remove(backend.raw_data_file_name)

    def test_get_measurement_outcome_distribution(self, backend):
        # Given
        circuit = Circuit([H(0), CNOT(0, 1), CNOT(1, 2)])
        n_samples = 1000

        try:
            # When
            distribution = backend.get_measurement_outcome_distribution(
                circuit, n_samples=n_samples
            )

            # Then
            assert isinstance(distribution, MeasurementOutcomeDistribution)
            assert distribution.get_number_of_subsystems() == 3
            assert distribution.distribution_dict[(0, 0, 0)] > 1 / 3
            assert distribution.distribution_dict[(1, 1, 1)] > 1 / 3
            assert backend.inner_backend.n_circuits_executed == 1
            assert backend.inner_backend.n_jobs_executed == 1
        finally:
            # Cleanup
            remove(backend.raw_data_file_name)

    def test_serialization_of_measurement_data_from_circuit(self, backend):
        try:
            # When
            backend.run_and_measure(Circuit([X(0), X(0)]), n_samples=10)
            with open(backend.raw_data_file_name) as f:
                data = load(f)

            # Then
            assert data["raw-data"][0]["device"] == "SymbolicSimulator"
            assert data["raw-data"][0]["number_of_shots"] == 10
            assert data["raw-data"][0]["data_type"] == "measurement"
            """Assert solutions are in the recorded data"""
            assert {"0": 10} == data["raw-data"][0]["counts"]
        finally:
            # Cleanup
            remove(backend.raw_data_file_name)

    def test_serialization_of_measurement_data_from_circuitset(self, backend):
        # Given
        circuitset = [Circuit([X(0), X(0)]), Circuit([X(0)])]
        n_samples = [10] * 2

        try:
            # When
            backend.run_batch_and_measure(circuitset, n_samples=n_samples)
            f = open(backend.raw_data_file_name)
            data = load(f)

            # Then
            for i in range(1):
                assert data["raw-data"][i]["device"] == "SymbolicSimulator"
                assert data["raw-data"][i]["number_of_shots"] == 10
                assert data["raw-data"][i]["data_type"] == "measurement"
            """Assert solutions are in the recorded data"""
            assert {"0": 10} == data["raw-data"][0]["counts"]
            assert {"1": 10} == data["raw-data"][1]["counts"]
        finally:
            # Cleanup
            remove(backend.raw_data_file_name)

    def test_serialization_of_bitstrings(self, backend):
        # Given
        backend.record_bitstrings = True

        try:
            # When
            backend.run_and_measure(Circuit([X(0), X(0)]), n_samples=10)
            f = open(backend.raw_data_file_name)
            data = load(f)

            # Then
            assert data["raw-data"][0]["device"] == "SymbolicSimulator"
            assert data["raw-data"][0]["number_of_shots"] == 10
            """Assert solutions are in the recorded data"""
            assert [[0] for _ in range(10)] == data["raw-data"][0]["bitstrings"]
        finally:
            # Cleanup
            remove(backend.raw_data_file_name)

    def test_serialization_of_measurement_outcome_distributions(self, backend):
        try:
            # When
            backend.get_measurement_outcome_distribution(
                Circuit([X(0), X(0)]), n_samples=10
            )
            with open(backend.raw_data_file_name) as f:
                data = load(f)

            # Then
            assert data["raw-data"][0]["device"] == "SymbolicSimulator"
            assert data["raw-data"][0]["number_of_shots"] == 10
            """Assert solutions are in the recorded data"""
            assert (
                "MeasurementOutcomeDistribution(input={(0,): 1.0})"
                == data["raw-data"][0]["distribution"]
            )
        finally:
            # Cleanup
            remove(backend.raw_data_file_name)

    def test_serialization_of_multiple_trackers_with_same_backend(self, backend):
        # Given
        backend_2 = MeasurementTrackingBackend(backend.inner_backend, "tracker_test_2")

        try:
            # When
            backend.run_and_measure(Circuit([X(0), X(0)]), n_samples=10)
            backend_2.run_and_measure(Circuit([X(0)]), n_samples=10)
            with open(backend.raw_data_file_name) as f:
                data = load(f)
            with open(backend_2.raw_data_file_name) as f_2:
                data_2 = load(f_2)

            # Then
            assert data["raw-data"][0]["device"] == "SymbolicSimulator"
            assert data["raw-data"][0]["number_of_shots"] == 10
            assert data_2["raw-data"][0]["device"] == "SymbolicSimulator"
            assert data_2["raw-data"][0]["number_of_shots"] == 10
            """Assert solutions are in the recorded data"""
            assert {"0": 10} == data["raw-data"][0]["counts"]
            assert {"1": 10} == data_2["raw-data"][0]["counts"]
            assert backend.inner_backend.n_circuits_executed == 2
            assert backend.inner_backend.n_jobs_executed == 2
        finally:
            # Cleanup
            remove(backend.raw_data_file_name)
            remove(backend_2.raw_data_file_name)

    def test_serialization_when_inner_backend_used_for_different_task(self, backend):
        # Given
        backend_2 = backend.inner_backend

        try:
            # When
            backend.run_and_measure(Circuit([X(0), X(0)]), n_samples=10)
            measurement = backend_2.run_and_measure(Circuit([X(0)]), n_samples=10)
            with open(backend.raw_data_file_name) as f:
                data = load(f)

            # Then
            assert data["raw-data"][0]["device"] == "SymbolicSimulator"
            assert data["raw-data"][0]["number_of_shots"] == 10
            """Assert solutions are in the recorded data"""
            assert {"0": 10} == data["raw-data"][0]["counts"]
            assert {"1": 10} == measurement.get_counts()
        finally:
            # Cleanup
            remove(backend.raw_data_file_name)



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
import numpy as np
import pytest
import sympy

from orquestra.quantum.circuits import RX, RY, RZ, XX, XY, YY, Circuit, H, I, X, Y, Z


class TestCreatingUnitaryFromCircuit:
    @pytest.fixture
    def cirq_unitaries(self):
        """
        Note: We decided to go with file-based approach after extracting cirq
        from orquestra-quantum and not being able to use `export_to_cirq` anymore.
        """
        path_to_array = (
            "/".join(__file__.split("/")[:-1]) + "/hardcoded_cirq_unitaries.npy"
        )
        return np.load(path_to_array, allow_pickle=True)

    @pytest.mark.parametrize(
        "circuit, unitary_index",
        [
            # Identity gates in some test cases below are used so that comparable
            # Cirq circuits have the same number of qubits as orquestra ones.
            (Circuit([RX(np.pi / 5)(0)]), 0),
            (Circuit([RY(np.pi / 2)(0), RX(np.pi / 5)(0)]), 1),
            (
                Circuit([I(1), I(2), I(3), I(4), RX(np.pi / 5)(0), XX(0.1)(5, 0)]),
                2,
            ),
            (
                Circuit(
                    [
                        XY(np.pi).controlled(1)(3, 1, 4),
                        RZ(0.1 * np.pi).controlled(2)(0, 2, 1),
                    ]
                ),
                3,
            ),
            (
                Circuit([H(1), YY(0.1).controlled(1)(0, 1, 2), X(2), Y(3), Z(4)]),
                4,
            ),
        ],
    )
    def test_without_free_params_gives_the_same_result_as_cirq(
        self, circuit, unitary_index, cirq_unitaries
    ):
        orquestra_unitary = circuit.to_unitary()

        assert isinstance(
            orquestra_unitary, np.ndarray
        ), "Unitary constructed from non-parameterized circuit is not a numpy array."

        np.testing.assert_array_almost_equal(
            orquestra_unitary, cirq_unitaries[unitary_index]
        )

    def test_commutes_with_parameter_substitution(self):
        theta, gamma = sympy.symbols("theta, gamma")
        circuit = Circuit(
            [RX(theta / 2)(0), X(1), RY(gamma / 4).controlled(1)(1, 2), YY(0.1)(0, 4)]
        )

        symbols_map = {theta: 0.1, gamma: 0.5}
        parameterized_unitary = circuit.to_unitary()
        unitary = circuit.bind(symbols_map).to_unitary()

        np.testing.assert_array_almost_equal(
            np.array(parameterized_unitary.subs(symbols_map), dtype=complex), unitary
        )



################################################################################
# © Copyright 2020-2022 Zapata Computing Inc.
################################################################################
import itertools
import math
import random

import numpy as np

from ..distributions import MeasurementOutcomeDistribution


def get_bars_and_stripes_target_distribution(
    nrows, ncols, fraction=1.0, method="zigzag"
):
    """Generates bars and stripes (BAS) data in zigzag pattern

    Args:
        nrows: number of rows in BAS dataset
        ncols: number of columns in BAS dataset
        fraction: maximum fraction of patterns to include
            (at least one pattern will always be included)
        method: the method to use to label the qubits
    Returns:
        Array of list of BAS pattern.
    """
    if method == "zigzag":
        data = bars_and_stripes_zigzag(nrows, ncols)
    else:
        raise RuntimeError(
            "Method: {} is not supported for generated a target distribution "
            "for bars and stripes".format(method)
        )

    # Remove patterns until left with a subset that has cardinality less than or equal
    # to the percentage * total number of patterns
    num_desired_patterns = int(len(data) * fraction)
    num_desired_patterns = max(num_desired_patterns, 1)
    data = random.sample(list(data), num_desired_patterns)

    distribution_dict = {}
    for pattern in data:
        bitstring = ""
        for qubit in pattern:
            bitstring += str(qubit)

        distribution_dict[bitstring] = 1.0

    return MeasurementOutcomeDistribution(distribution_dict)


# Generate BAS with specified rows and columns in zigzag pattern
def bars_and_stripes_zigzag(nrows, ncols):
    """Generates bars and stripes data in zigzag pattern

    Args:
        nrows: number of rows in BAS dataset
        ncols: number of columns in BAS dataset
    Returns:
        Array of list of BAS pattern.
    """

    data = []

    for h in itertools.product([0, 1], repeat=ncols):
        pic = np.repeat([h], nrows, 0)
        data.append(pic.ravel().tolist())

    for h in itertools.product([0, 1], repeat=nrows):
        pic = np.repeat([h], ncols, 1)
        data.append(pic.ravel().tolist())

    data = np.unique(np.asarray(data), axis=0)

    return data


def get_num_bars_and_stripes_patterns(nrows, ncols) -> int:
    """Get the number of bars and stripes patters for a 2-dimensional grid.

    Args:
        nrows (int): number of rows in BAS dataset
        ncols (int): number of columns in BAS dataset
    Returns:
        (int): number of bars and stripes patterns
    """
    # Always have all blank and all filled
    num_patterns = 2

    for dimension in [nrows, ncols]:
        for num_choices in range(1, dimension):
            # nCr = n! / (n-r)! * r!
            num_patterns += math.factorial(dimension) // (
                math.factorial(dimension - num_choices) * math.factorial(num_choices)
            )

    return num_patterns



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
import json
from typing import Any, Dict, Iterable, List, Optional, Union, cast

import numpy as np

from ..typing import AnyPath, LoadSource
from ..utils import convert_array_to_dict, convert_dict_to_array, ensure_open
from .parities import Parities


class ExpectationValues:
    """A class representing expectation values of operators.
    For more context on how it is being used, please see the docstring of
    EstimateExpectationValues Protocol in interfaces/estimation.py.

    Args:
        values: The expectation values of a set of terms in an ising operator.
        correlations: The expectation values of pairwise products of operators.
            Contains an NxN array for each frame, where N is the number of
            operators in that frame.
        estimator_covariances: The (estimated) covariances between estimates of
            expectation values of pairs of operators. Contains an NxN array for
            each frame, where N is the number of operators in that frame. Note
            that for direct sampling, the covariance between estimates of
            expectation values is equal to the population covariance divided by
            the number of samples.

    Attributes:
        values: See Args.
        correlations: See Args.
        estimator_covariances: See Args.
    """

    def __init__(
        self,
        values: np.ndarray,
        correlations: Optional[List[np.ndarray]] = None,
        estimator_covariances: Optional[List[np.ndarray]] = None,
    ):
        self.values = values
        self.correlations = correlations
        self.estimator_covariances = estimator_covariances

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dictionary"""

        data: Dict[str, Any] = {
            "frames": [],
        }  # what is "frames" for?

        data["expectation_values"] = convert_array_to_dict(self.values)

        if self.correlations:
            data["correlations"] = []
            for correlation_matrix in self.correlations:
                data["correlations"].append(convert_array_to_dict(correlation_matrix))

        if self.estimator_covariances:
            data["estimator_covariances"] = []
            for covariance_matrix in self.estimator_covariances:
                data["estimator_covariances"].append(
                    convert_array_to_dict(covariance_matrix)
                )

        return data

    @classmethod
    def from_dict(cls, dictionary: dict) -> "ExpectationValues":
        """Create an ExpectationValues object from a dictionary."""

        expectation_values = convert_dict_to_array(dictionary["expectation_values"])
        correlations: Optional[List] = None
        if dictionary.get("correlations"):
            correlations = []
            for correlation_matrix in cast(Iterable, dictionary.get("correlations")):
                correlations.append(convert_dict_to_array(correlation_matrix))

        estimator_covariances: Union[List, None] = None
        if dictionary.get("estimator_covariances"):
            estimator_covariances = []
            for covariance_matrix in cast(
                Iterable, dictionary.get("estimator_covariances")
            ):
                estimator_covariances.append(convert_dict_to_array(covariance_matrix))

        return cls(expectation_values, correlations, estimator_covariances)

    def __eq__(self, __o: object) -> bool:
        return self.__dict__ == __o.__dict__


def save_expectation_values(
    expectation_values: ExpectationValues, filename: AnyPath
) -> None:
    """Save expectation values to a file.

    Args:
        expectation_values (ExpectationValues): the expectation values to save
        file (str or file-like object): the name of the file, or a file-like object
    """
    dictionary = expectation_values.to_dict()

    with open(filename, "w") as f:
        f.write(json.dumps(dictionary, indent=2))


def load_expectation_values(file: LoadSource) -> ExpectationValues:
    """Load an array from a file.

    Args:
        file (str or file-like object): the name of the file, or a file-like object.

    Returns:
        array (numpy.array): the array
    """

    with ensure_open(file) as f:
        data = json.load(f)

    return ExpectationValues.from_dict(data)


def expectation_values_to_real(
    expectation_values: ExpectationValues,
) -> ExpectationValues:
    """Remove the imaginary parts of the expectation values

    Args:
        expectation_values (orquestra.quantum.measurement.ExpectationValues object)
    Returns:
        expectation_values (orquestra.quantum.measurement.ExpectationValues object)
    """
    values = []
    for i, value in enumerate(expectation_values.values):
        if isinstance(value, complex):
            value = value.real
        values.append(value)
    expectation_values.values = np.array(values)
    if expectation_values.correlations:
        for i, value in enumerate(expectation_values.correlations):
            if isinstance(value, complex):
                value = value.real
            expectation_values.correlations[i] = value
    return expectation_values


def concatenate_expectation_values(
    expectation_values_set: Iterable[ExpectationValues],
) -> ExpectationValues:
    """Concatenates a set of expectation values objects.

    Args:
        expectation_values_set: The expectation values objects to be concatenated.

    Returns:
        The combined expectation values.
    """

    combined_expectation_values = ExpectationValues(np.zeros(0))

    for expectation_values in expectation_values_set:
        combined_expectation_values.values = np.concatenate(
            (combined_expectation_values.values, expectation_values.values)
        )
        if expectation_values.correlations:
            if not combined_expectation_values.correlations:
                combined_expectation_values.correlations = []
            combined_expectation_values.correlations += expectation_values.correlations
        if expectation_values.estimator_covariances:
            if not combined_expectation_values.estimator_covariances:
                combined_expectation_values.estimator_covariances = []
            combined_expectation_values.estimator_covariances += (
                expectation_values.estimator_covariances
            )

    return combined_expectation_values


def get_expectation_values_from_parities(parities: Parities) -> ExpectationValues:
    """Get the expectation values of a set of operators (with precisions) from a set of
    samples (with even/odd parities) for them.

    Args:
        parities: Contains the number of samples with even and odd parities for each
            operator.

    Returns:
        Expectation values of the operators and the associated precisions.
    """
    values = []
    estimator_covariances = []

    for i in range(len(parities.values)):
        N0 = parities.values[i][0]
        N1 = parities.values[i][1]
        N = N0 + N1
        if N == 0:
            raise ValueError("There must be at least one sample for each operator")

        p = N0 / N
        value = 2.0 * p - 1.0

        # If there are enough samples and the probability of getting a sample with even
        # parity is not close to 0 or 1, then we can use p=N0/N to approximate this
        # probability and plug it into the formula for the precision.
        if N >= 100 and p >= 0.1 and p <= 0.9:
            precision = 2.0 * np.sqrt(p * (1.0 - p)) / np.sqrt(N)
        else:
            # Otherwise, p=N0/N may be not a good approximation of this probability.
            # So we use an upper bound on the precision instead.
            precision = 1.0 / np.sqrt(N)

        values.append(value)
        estimator_covariances.append(np.array([[precision**2.0]]))

    return ExpectationValues(
        values=np.array(values), estimator_covariances=estimator_covariances
    )



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
"""Test cases for _gates module."""
from unittest.mock import Mock

import numpy as np
import pytest
import sympy

from orquestra.quantum.circuits import _builtin_gates, _gates
from orquestra.quantum.circuits._gates import (
    Exponential,
    GateOperation,
    MatrixFactoryGate,
    Power,
)

GATES_REPRESENTATIVES = [
    _builtin_gates.X,
    _builtin_gates.Y,
    _builtin_gates.Z,
    _builtin_gates.T,
    _builtin_gates.H,
    _builtin_gates.I,
    _builtin_gates.RX(sympy.Symbol("theta")),
    _builtin_gates.RY(0.5),
    _builtin_gates.RZ(0),
    _builtin_gates.PHASE(sympy.pi / 5),
    _builtin_gates.U3(np.pi, sympy.pi / 2, sympy.Symbol("x")),
    _builtin_gates.GPi(sympy.Symbol("theta")),
    _builtin_gates.GPi2(sympy.Symbol("theta")),
    _builtin_gates.CZ,
    _builtin_gates.CNOT,
    _builtin_gates.SWAP,
    _builtin_gates.ISWAP,
    _builtin_gates.XX(sympy.cos(sympy.Symbol("phi"))),
    _builtin_gates.YY(sympy.pi),
    _builtin_gates.ZZ(sympy.Symbol("x") + sympy.Symbol("y")),
    _builtin_gates.CPHASE(1.5),
    _builtin_gates.MS(sympy.Symbol("theta"), np.pi / 2),
]

POWER_GATE_EXPONENTS = [-2.0, 0, 0.5, 1.0]


def example_one_qubit_matrix_factory(a, b):
    return sympy.Matrix([[a, b], [b, a]])


def example_two_qubit_matrix_factory(a, b, c):
    return sympy.Matrix([[a, 0, 0, 0], [0, b, 0, 0], [0, 0, c, 0], [0, 0, 0, 1]])


class TestMatrixFactoryGate:
    @pytest.mark.parametrize(
        "params, factory, num_qubits",
        [
            ((0.5, sympy.Symbol("theta")), example_one_qubit_matrix_factory, 1),
            (
                (sympy.Symbol("alpha"), sympy.Symbol("beta"), 1),
                example_two_qubit_matrix_factory,
                2,
            ),
        ],
    )
    def test_constructs_its_matrix_by_calling_factory_with_bound_parameter(
        self, params, factory, num_qubits
    ):
        wrapped_factory = Mock(wraps=factory)
        gate = MatrixFactoryGate("U", wrapped_factory, params, num_qubits)
        assert gate.matrix == factory(*params)
        wrapped_factory.assert_called_once_with(*params)

    def test_binding_parameters_creates_new_instance_with_substituted_free_params(self):
        gamma, theta, x, y = sympy.symbols("gamma, theta, x, y")
        params = (theta, x + y)
        gate = MatrixFactoryGate("U", example_one_qubit_matrix_factory, params, 1)

        new_gate = gate.bind({theta: 0.5, x: gamma, y: 3})

        assert new_gate.name == gate.name
        assert new_gate.matrix_factory == gate.matrix_factory
        assert new_gate.num_qubits == gate.num_qubits
        assert new_gate.params == (0.5, gamma + 3)

    def test_binding_parameters_with_symbol_outside_of_free_symbols_does_not_raise(
        self,
    ):
        gamma, theta = sympy.symbols("gamma, theta")
        params = (theta, 2 * theta)
        gate = MatrixFactoryGate("U", example_one_qubit_matrix_factory, params, 1)

        new_gate = gate.bind({gamma: 0.5, theta: 1})

        assert new_gate.params == (1, 2)

    def test_binding_parameters_does_not_change_parameters_without_free_symbols(self):
        theta = sympy.Symbol("theta")
        gate = MatrixFactoryGate("V", example_one_qubit_matrix_factory, (1, 2), 1)

        new_gate = gate.bind({theta: 5.0})

        assert new_gate.params == (1, 2)

    def test_replace_parameters_correctly_gives_instance_with_correctly_set_parameters(
        self,
    ):
        theta = sympy.Symbol("theta")
        gate = MatrixFactoryGate("V", example_one_qubit_matrix_factory, (1, 2), 1)

        new_gate = gate.replace_params((theta, 0.5))

        assert new_gate == MatrixFactoryGate(
            "V", example_one_qubit_matrix_factory, (theta, 0.5), 1
        )

    def test_daggers_matrix_is_adjoint_of_original_gates_matrix(self):
        gate = MatrixFactoryGate("V", example_one_qubit_matrix_factory, (1, 2), 1)
        assert gate.dagger.matrix == gate.matrix.adjoint()

    def test_matrix_exponential_is_exponential_of_original_gates_matrix(self):
        gate = MatrixFactoryGate("V", example_one_qubit_matrix_factory, (1, 2), 1)
        assert gate.exp.matrix == gate.matrix.exp()

    def test_dagger_has_the_same_params_and_num_qubits_as_wrapped_gate(self):
        gate = MatrixFactoryGate(
            "U", example_two_qubit_matrix_factory, (0.5, 0.1, sympy.Symbol("a")), 2
        )
        assert gate.dagger.num_qubits == gate.num_qubits
        assert gate.dagger.params == gate.params

    def test_dagger_of_hermitian_gate_is_the_same_gate(self):
        gate = MatrixFactoryGate(
            "V", example_one_qubit_matrix_factory, (1, 0), 1, is_hermitian=True
        )
        assert gate.dagger is gate

    def test_power_of_dagger_is_dagger_wrapped_by_power(self):
        gate = MatrixFactoryGate("V", example_one_qubit_matrix_factory, (1, 0), 1)
        assert gate.dagger.power(0.5) == Power(gate.dagger, 0.5)

    def test_exponential_of_dagger_is_dagger_wrapped_by_exponential(self):
        gate = MatrixFactoryGate("V", example_one_qubit_matrix_factory, (1, 0), 1)
        assert gate.dagger.exp == Exponential(gate.dagger)

    def test_binding_gates_in_dagger_is_propagated_to_wrapped_gate(self):
        theta = sympy.Symbol("theta")
        gate = MatrixFactoryGate("V", example_one_qubit_matrix_factory, (theta, 0), 1)

        assert gate.dagger.bind({theta: 0.5}) == gate.bind({theta: 0.5}).dagger

    def test_dagger_of_dagger_is_the_same_as_original_gate(self):
        gate = MatrixFactoryGate("V", example_one_qubit_matrix_factory, (1, 0), 1)
        assert gate.dagger.dagger is gate

    def test_applying_dagger_and_replacing_parameters_commutes(self):
        gate = MatrixFactoryGate("V", example_one_qubit_matrix_factory, (1, 0), 1)
        new_params = (sympy.Symbol("theta"), 4.2)
        assert (
            gate.dagger.replace_params(new_params)
            == gate.replace_params(new_params).dagger
        )

    def test_applying_gate_returns_operation_with_correct_gate_and_indices(self):
        theta = sympy.Symbol("theta")
        gamma = sympy.Symbol("gamma")
        gate = MatrixFactoryGate(
            "A", example_two_qubit_matrix_factory, (theta, gamma, 42), 2
        )
        operation = gate(4, 1)

        assert operation.gate == gate
        assert operation.qubit_indices == (4, 1)

    def test_str_on_dagger_gives_correct_representation(self):
        gate = MatrixFactoryGate("V", example_one_qubit_matrix_factory, (1, 0), 1)
        assert str(gate.dagger) == "V†(1, 0)"

    def test_wrapping_two_times_gives_correct_string(self):
        phi = sympy.Symbol("phi")
        op = MatrixFactoryGate("U", example_one_qubit_matrix_factory, (phi, 1), 1)
        T_op = _builtin_gates.T  # need a gate without params for exponentiation

        # assert all permutations produce correct string
        assert str(op.controlled(2).dagger(0)) == "c-c-U†(phi, 1)(0)"
        assert str(T_op.controlled(2).exp(0)) == "exp^{c-c-T}(0)"
        assert str(T_op.controlled(2).power(3)(0)) == "c-c-T^3(0)"
        assert str(op.controlled(2).controlled(2)(0)) == "c-c-c-c-U(phi, 1)(0)"

        assert str(op.dagger.controlled(2)(0)) == "c-c-U†(phi, 1)(0)"
        assert str(T_op.dagger.exp(0)) == "exp^T†(0)"
        assert str(T_op.dagger.power(3)(0)) == "T†^3(0)"
        assert str(op.dagger.dagger(0)) == "U(phi, 1)(0)"

        assert str(T_op.exp.dagger(0)) == "exp^T†(0)"
        assert str(T_op.exp.controlled(2)(0)) == "c-c-exp^T(0)"
        assert str(T_op.exp.power(3)(0)) == "{exp^T}^3(0)"
        assert str(T_op.exp.exp(0)) == "exp^exp^T(0)"

        assert str(T_op.power(3).dagger(0)) == "T†^3(0)"
        assert str(T_op.power(3).controlled(2)(0)) == "c-c-T^3(0)"
        assert str(T_op.power(3).exp(0)) == "exp^{T^3}(0)"
        assert str(T_op.power(3).power(3)(0)) == "T^3^3(0)"

    def test_wrapping_three_times_gives_correct_string(self):
        T_op = _builtin_gates.T  # need a gate without params for exponentiation

        # assert some permutations because there are too many to test exhaustively
        assert str(T_op.controlled(2).dagger.exp(0)) == "exp^{c-c-T†}(0)"
        assert str(T_op.dagger.controlled(2).power(3)(0)) == "c-c-T†^3(0)"
        assert str(T_op.exp.controlled(2).exp(0)) == "exp^{c-c-exp^T}(0)"


@pytest.mark.parametrize("gate", GATES_REPRESENTATIVES)
class TestControlledGate:
    def test_num_qubits_equal_to_wrapped_gates_num_qubits_plus_num_controlled_qubits(
        self, gate
    ):
        assert gate.controlled(3).num_qubits == gate.num_qubits + 3

    def test_has_matrix_with_eye_and_wrapped_gates_matrix_as_bottom_left_block(
        self, gate
    ):
        controlled_gate = gate.controlled(2)
        n = gate.matrix.shape[0]
        assert gate.matrix.shape[1] == n
        assert controlled_gate.matrix[0:-n, 0:-n] == sympy.eye(
            2**controlled_gate.num_qubits - n
        )
        assert controlled_gate.matrix[-n:, -n:] == gate.matrix

    def test_controlled_of_controlled_gate_has_summed_number_of_control_qubits(
        self, gate
    ):
        controlled_gate = gate.controlled(2)
        double_controlled_gate = controlled_gate.controlled(3)

        assert double_controlled_gate.wrapped_gate == gate
        assert double_controlled_gate.num_qubits == gate.num_qubits + 2 + 3
        assert double_controlled_gate.num_control_qubits == 2 + 3
        assert double_controlled_gate.matrix.shape == 2 * (
            2 ** (gate.num_qubits + 2 + 3),
        )

    def test_has_the_same_parameters_as_wrapped_gate(self, gate):
        controlled_gate = gate.controlled(4)

        assert controlled_gate.params == gate.params

    def test_dagger_of_controlled_gate_is_controlled_gate_wrapping_dagger(self, gate):
        controlled_gate = gate.controlled(4)

        assert controlled_gate.dagger == gate.dagger.controlled(4)

    def test_exp_of_controlled_gate_is_not_controlled_gate_wrapping_exp(self, gate):
        if len(gate.free_symbols) == 0:
            controlled_gate = gate.controlled(2)
            assert controlled_gate.exp != gate.exp.controlled(2)

    def test_power_of_controlled_gate_is_controlled_gate_wrapping_power(self, gate):
        if len(gate.free_symbols) == 0:
            controlled_gate = gate.controlled(2)
            assert controlled_gate.power(0.5) == gate.power(0.5).controlled(2)

    def test_binding_parameters_in_control_gate_is_propagated_to_wrapped_gate(
        self, gate
    ):
        controlled_gate = gate.controlled(2)
        symbols_map = {sympy.Symbol("theta"): 0.5, sympy.Symbol("x"): 3}
        assert controlled_gate.bind(symbols_map) == gate.bind(symbols_map).controlled(2)

    def test_constructing_controlled_gate_and_replacing_parameters_commute(self, gate):
        controlled_gate = gate.controlled(2)
        new_params = tuple(3 * param for param in controlled_gate.params)

        assert controlled_gate.replace_params(new_params) == gate.replace_params(
            new_params
        ).controlled(2)

    def test_constructing_controlled_gate_with_zero_control_raises_error(self, gate):
        with pytest.raises(ValueError):
            gate.controlled(0)

    def test_str_gives_correct_string_for_one_control(self, gate):
        controlled_gate = gate.controlled(1)
        assert str(controlled_gate) == "c-" + str(gate)

    def test_str_gives_correct_string_for_multiple_controls(self, gate):
        controlled_gate_2 = gate.controlled(2)
        controlled_gate_5 = gate.controlled(5)
        assert str(controlled_gate_2) == "c-" * 2 + str(gate)
        assert str(controlled_gate_5) == "c-" * 5 + str(gate)

    def test_str_gives_correct_string_for_stacking_controls(self, gate):
        double_controlled_gate = gate.controlled(1).controlled(1)
        assert str(double_controlled_gate) == "c-" * 2 + str(gate)


@pytest.mark.parametrize("gate", GATES_REPRESENTATIVES)
@pytest.mark.parametrize("exponent", POWER_GATE_EXPONENTS)
class TestPowerGate:
    def test_constructing_a_power_gate_with_free_symbols_raises_error(
        self, gate, exponent
    ):
        if len(gate.free_symbols) > 0:
            with pytest.raises(ValueError):
                gate.power(exponent)

    def test_power_gate_naming_scheme(self, gate, exponent):
        if len(gate.free_symbols) == 0:
            power_gate = gate.power(exponent)
            assert power_gate.name == f"{gate.name}{_gates.POWER_GATE_SYMBOL}{exponent}"

    def test_has_same_parameters_as_wrapped_gate(self, gate, exponent):
        if len(gate.free_symbols) == 0:
            assert gate.power(exponent).params == gate.params

    def test_has_same_free_symbols_as_wrapped_gate(self, gate, exponent):
        if len(gate.free_symbols) == 0:
            assert gate.power(exponent).free_symbols == gate.free_symbols

    def test_has_same_number_of_qubits_as_wrapped_gate(self, gate, exponent):
        if len(gate.free_symbols) == 0:
            assert gate.power(exponent).num_qubits == gate.num_qubits

    def test_has_matrix_equal_to_wrapped_gate_matrix_exponentiated(
        self, gate, exponent
    ):
        if len(gate.free_symbols) == 0:
            wrapped_gate_matrix_exponentiated = gate.matrix**exponent
            assert gate.power(exponent).matrix == wrapped_gate_matrix_exponentiated

    def test_dagger_of_power_gate_power_gate_of_dagger(self, gate, exponent):
        if len(gate.free_symbols) == 0:
            power_gate = gate.power(exponent)
            assert power_gate.dagger == gate.dagger.power(exponent)

    def test_creating_power_gate_from_power_gate(self, gate, exponent):
        if len(gate.free_symbols) == 0:
            power_gate = gate.power(exponent)
            powered_power_gate = power_gate.power(exponent)
            assert powered_power_gate.matrix == power_gate.matrix**exponent

    def test_parameter_binding_not_implemented_for_power_gates(self, gate, exponent):
        if len(gate.free_symbols) == 0:
            power_gate = gate.power(exponent)
            symbols_map = {sympy.Symbol("theta"): 0.5, sympy.Symbol("x"): 3}
            with pytest.raises(NotImplementedError):
                power_gate.bind(symbols_map)

    def test_constructing_power_gate_and_replacing_parameters_commute(
        self, gate, exponent
    ):
        if len(gate.free_symbols) == 0:
            power_gate = gate.power(exponent)
            new_params = tuple(3 * param for param in power_gate.params)

            assert power_gate.replace_params(new_params) == gate.replace_params(
                new_params
            ).power(exponent)

    def test_str_gives_correct_string(self, gate, exponent):
        if len(gate.free_symbols) == 0:
            power_gate = gate.power(exponent)
            correct_str = str(gate) + f"{_gates.POWER_GATE_SYMBOL}{exponent}"
            assert str(power_gate) == correct_str


@pytest.mark.parametrize("gate", GATES_REPRESENTATIVES[:10])
class TestGateExponential:
    def test_constructing_a_gate_exponential_with_free_symbols_raises_error(self, gate):
        if len(gate.free_symbols) > 0:
            with pytest.raises(ValueError):
                gate.exp

    def test_gate_exponential_naming_scheme(self, gate):
        if len(gate.free_symbols) == 0:
            gate_exponential = gate.exp
            assert gate_exponential.name == _gates.EXPONENTIAL_GATE_NAME

    def test_has_same_parameters_as_wrapped_gate(self, gate):
        if len(gate.free_symbols) == 0:
            assert gate.exp.params == gate.params

    def test_has_same_free_symbols_as_wrapped_gate(self, gate):
        if len(gate.free_symbols) == 0:
            assert gate.exp.free_symbols == gate.free_symbols

    def test_has_same_number_of_qubits_as_wrapped_gate(self, gate):
        if len(gate.free_symbols) == 0:
            assert gate.exp.num_qubits == gate.num_qubits

    def test_matrix_exponential_equal_to_wrapped_gate_matrix_exponential(self, gate):
        if len(gate.free_symbols) == 0 and gate.name != "T":
            wrapped_gate_matrix_exponential = gate.matrix.exp()
            assert gate.exp.matrix == wrapped_gate_matrix_exponential

    def test_dagger_of_gate_exponential_gate_exponential_of_dagger(self, gate):
        if len(gate.free_symbols) == 0 and gate.name != "T":
            assert gate.exp.dagger == gate.dagger.exp

    def test_power_of_gate_exponential_gate_exponential_of_power(self, gate):
        if len(gate.free_symbols) == 0 and gate.name != "T":
            assert gate.exp.power(2.0) != gate.power(2.0).exp

    def test_parameter_binding_not_implemented_for_gate_exponential(self, gate):
        if len(gate.free_symbols) == 0:
            gate_exponential = gate.exp
            symbols_map = {sympy.Symbol("theta"): 0.5, sympy.Symbol("x"): 3}
            with pytest.raises(NotImplementedError):
                gate_exponential.bind(symbols_map)

    def test_constructing_gate_exponential_and_replacing_parameters_commute(self, gate):
        if len(gate.free_symbols) == 0:
            gate_exponential = gate.exp
            new_params = tuple(3 * param for param in gate_exponential.params)

            assert (
                gate_exponential.replace_params(new_params)
                == gate.replace_params(new_params).exp
            )

    def test_str_gives_correct_string(self, gate):
        if len(gate.free_symbols) == 0:
            assert str(gate.exp) == "exp^" + str(gate)


@pytest.mark.parametrize("gate", GATES_REPRESENTATIVES)
class TestGateOperation:
    def test_bound_symbols_are_not_present_in_gate_parameters(self, gate):
        op = GateOperation(gate, tuple(range(gate.num_qubits)))
        symbols_map = {sympy.Symbol("phi"): 0.5, sympy.Symbol("y"): 1.1}
        assert all(
            symbol not in sympy.sympify(param).atoms(sympy.Symbol)
            for symbol in symbols_map
            for param in op.bind(symbols_map).params
        )

    def test_replacing_parameters_constructs_operation_of_gate_with_new_parameters(
        self, gate
    ):
        op = GateOperation(gate, tuple(range(gate.num_qubits)))
        new_params = tuple(-1 * param for param in op.params)

        assert op.replace_params(new_params).params == new_params

    def test_free_symbols_of_gate_operation_are_the_same_as_the_ones_in_wrapped_gate(
        self, gate
    ):
        op = GateOperation(gate, tuple(range(gate.num_qubits)))
        assert op.free_symbols == gate.free_symbols

    def test_cannot_be_applied_to_vector_of_not_power_of_two_length(self, gate):
        state_vector = np.array([0.1 for _ in range(2**gate.num_qubits + 1)])
        with pytest.raises(ValueError):
            GateOperation(gate, tuple(range(gate.num_qubits))).apply(state_vector)



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
"""Test case prototypes of instances of the EstimateExpectationValues protocol
that can be used in other projects.

Note that this file won't be executed on its own by pytest.
You need to define your own test cases that import the ones defined here.
Here is an example of how you would do that:

.. code:: python

    from orquestra.quantum.api.estimator_contract import ESTIMATOR_CONTRACTS

    @pytest.mark.parametrize("contract", ESTIMATOR_CONTRACTS)
    def test_estimator_contract(contract):
        estimator = CvarEstimator(alpha=0.2)
        assert contract(estimator)
"""

import numpy as np

from ..api.estimation import EstimateExpectationValues, EstimationTask
from ..circuits import RX, RY, RZ, Circuit, H
from ..operators import PauliTerm
from ..runners.symbolic_simulator import SymbolicSimulator

_backend = SymbolicSimulator(seed=1997)

_estimation_tasks = [
    EstimationTask(PauliTerm("Z0"), Circuit([H(0)]), 10000),
    EstimationTask(
        PauliTerm("Z0") + PauliTerm("Z1") + PauliTerm("Z2"),
        Circuit([H(0), RX(np.pi / 3)(0), H(2)]),
        10000,
    ),
    EstimationTask(
        PauliTerm("Z0") + PauliTerm("Z1", 4),
        Circuit(
            [
                RX(np.pi)(0),
                RY(0.12)(1),
                RZ(np.pi / 3)(1),
                RY(1.9213)(0),
            ]
        ),
        10000,
    ),
]


def _validate_each_task_returns_one_expecation_value(
    estimator: EstimateExpectationValues,
):
    # When
    expectation_values = estimator(
        runner=_backend,
        estimation_tasks=_estimation_tasks,
    )

    # Then
    return len(expectation_values) == len(_estimation_tasks)


def _validate_order_of_outputs_matches_order_of_inputs(
    estimator: EstimateExpectationValues,
):
    expectation_values = estimator(
        runner=_backend,
        estimation_tasks=_estimation_tasks,
    )

    return all(
        [
            np.allclose(
                expectation_values[i].values,
                estimator(
                    runner=_backend,
                    estimation_tasks=[task],
                )[0].values,
                rtol=0.1,  # 10% tolerance
            )
            for i, task in enumerate(_estimation_tasks)
        ]
    )


def _validate_expectation_value_includes_coefficients(
    estimator: EstimateExpectationValues,
):
    term_coefficient = 19.971997
    estimation_tasks = [
        EstimationTask(PauliTerm("Z0"), Circuit([RX(np.pi / 3)(0)]), 10000),
        EstimationTask(
            PauliTerm("Z0", term_coefficient), Circuit([RX(np.pi / 3)(0)]), 10000
        ),
    ]

    expectation_values = estimator(
        runner=_backend,
        estimation_tasks=estimation_tasks,
    )

    return np.allclose(
        expectation_values[0].values,
        expectation_values[1].values / term_coefficient,
        rtol=0.1,  # 10% tolerance
    )


def _validate_constant_terms_are_included_in_output(
    estimator: EstimateExpectationValues,
):
    estimation_tasks = [
        EstimationTask(PauliTerm("Z0"), Circuit([H(0)]), 10000),
        EstimationTask(
            PauliTerm("Z0") + PauliTerm("I0", 19.971997),
            Circuit([H(0)]),
            10000,
        ),
    ]

    expectation_values = estimator(
        runner=_backend,
        estimation_tasks=estimation_tasks,
    )

    return not np.array_equal(
        expectation_values[0].values, expectation_values[1].values
    )


ESTIMATOR_CONTRACTS = [
    _validate_each_task_returns_one_expecation_value,
    _validate_order_of_outputs_matches_order_of_inputs,
    _validate_expectation_value_includes_coefficients,
    _validate_constant_terms_are_included_in_output,
]



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
"""The WavefunctionSimulator protocol and ABC for implementing it."""
from abc import abstractmethod
from typing import Optional, Protocol

import numpy as np

from ..circuits import Circuit, GateOperation, Operation, split_circuit
from ..distributions import (
    MeasurementOutcomeDistribution,
    create_bitstring_distribution_from_probability_distribution,
)
from ..measurements import Measurements
from ..operators import PauliRepresentation, get_expectation_value
from ..typing import StateVector
from ..wavefunction import Wavefunction, sample_from_wavefunction
from .circuit_runner import BaseCircuitRunner, CircuitRunner


class WavefunctionSimulator(CircuitRunner, Protocol):
    """The protocol for objects able to compute the wavefunction."""

    def get_wavefunction(
        self, circuit: Circuit, initial_state: Optional[StateVector] = None
    ) -> Wavefunction:
        """Get a wavefunction of a circuit starting from a given initial state.

        Args:
            circuit: circuit whose wavefunction is to be computed.
            initial_state: state used as an input to the `circuit`. If not provided,
              `|0...0>` will be used.
        Returns:
            Wavefunction comprising 2 ** n amplitudes, where n is number of qubits in
            `circuit`.
        """

    def get_exact_expectation_values(
        self, circuit: Circuit, operator: PauliRepresentation
    ) -> float:
        """Get an exact expectation value of an operator.

        Args:
            circuit: circuit constructing state for which expectation value
              is to be computed.
            operator: operator of which expectation value should be computed.
        Returns:
            An expectation value as a single floating point number.
        """


class BaseWavefunctionSimulator(BaseCircuitRunner, WavefunctionSimulator):
    """ABC for implementing simple wavefunction simulators.

    This ABC is build around _get_wavefunction_from_native_circuit
    method. In general, most simulators wrap some third-party resource (
    a library, service, API etc.), which can only consume circuits comprising
    operations from a given set. Such operations are called "native" to
    the given simulator, whereas other operations are called "nonnative".

    The idea of simulating arbitrary circuit is thus as follows:

    - split circuit into alternating consecutive parts of only native and
      only "nonnative" operations.

    - start with some initial state

    - for each part:

      - if it is native, run it via third-party resource, save the new
        statevector

      - if it is nonnative, apply each operation in the sequence using
        operation.apply(previous_statevector). Save the new statevector.

      Last saved statevector is the wavefunction of the total circuit.

      For this to work, subclasses of this ABC should implement
      _get_wavefunction_from_the_native_circuit method.

    Note:
        Since this class inherits all the limitations of BaseCircuitRunner.
        The _run_and_measure function is implemented via sampling from the
        wavefunction. Care must be taken when using third-party service that
        implements more sophisticated/more performant sampling method not
        involving direct computation of the whole wavefunction. In such cases,
        using BaseWavefunctionSimulator ABC will likely result in huge
        performance hit.
    """

    def __init__(self, *, seed: Optional[int] = None):
        super().__init__()
        self.seed = seed

    def run_and_measure(self, circuit: Circuit, n_samples: int) -> Measurements:
        """Run circuit given number of times and return obtained measurements.

        See docstrings of CircuitRunner protocol for exact description of
        parameters.

        Note:
            For BaseWavefunctionSimulator, running a circuit comprising computing
            its wavefunction and then sampling from the probability distribution
            that it defines.
        """
        if n_samples <= 0:
            raise ValueError(f"Number of samples has to be positive, got {n_samples}")
        result = self._run_and_measure(circuit, n_samples)
        return result

    def _run_and_measure(self, circuit: Circuit, n_samples: int) -> Measurements:
        if circuit.free_symbols:
            raise ValueError("Cannot sample from circuit with unbound free symbols")
        wavefunction = self.get_wavefunction(circuit)
        return Measurements(
            sample_from_wavefunction(wavefunction, n_samples, self.seed)
        )

    def get_wavefunction(
        self, circuit: Circuit, initial_state: Optional[StateVector] = None
    ) -> Wavefunction:
        """Get a wavefunction of a circuit starting from a given initial state.

        See docstrings of WavefunctionSimulator protocol for exact description of
        parameters.
        """
        state: StateVector
        if initial_state is None:
            state = np.zeros(2**circuit.n_qubits)
            state[0] = 1
        else:
            state = initial_state

        for is_supported, subcircuit in split_circuit(
            circuit, self.is_natively_supported
        ):
            # Native subcircuits are passed through to the underlying simulator.
            # They also count towards number of circuits and number of jobs run.
            self._n_jobs_executed += 1
            if is_supported:
                self._n_circuits_executed += 1
                state = self._get_wavefunction_from_native_circuit(subcircuit, state)
            else:
                for operation in subcircuit.operations:
                    state = operation.apply(state)

        return Wavefunction(state)

    @abstractmethod
    def _get_wavefunction_from_native_circuit(
        self, circuit: Circuit, initial_state: StateVector
    ) -> StateVector:
        """Get a wavefunction of a native circuit starting from given initial state.

        This is the only mandatory method to be implemented by classes extending
        this ABC.

        Args:
            circuit: circuit whose wavefunction is to be computed. This method might
              assume that it comprises only operations native to this simulator.
            initial_state: state used as an input to the `circuit`. If not provided,
              |0...0> will be used.
        Returns:
            The computed wavefunction.
        """

    def get_exact_expectation_values(
        self, circuit: Circuit, operator: PauliRepresentation
    ) -> float:
        """Get an exact expectation value of an operator.

        See docstrings of WavefunctionSimulator protocol for exact description of
        parameters.
        """
        wavefunction = self.get_wavefunction(circuit)
        # Casting to real, because any non-zero imaginary part must mean some
        # numerical inaccuracy.
        return get_expectation_value(operator, wavefunction).real

    def is_natively_supported(self, operation: Operation) -> bool:
        """Determine if given operation is natively supported by this simulator.

        By default, operation is natively supported iff it is a GateOperation.
        However, this doesn't have to be true for every simulator.
        If the set of natively supported operations is different for
        some simulator, this method should be changed accordingly.

        Args:
            operation: operation to be checked.
        Returns:
              True if `operation` is natively supported and False otherwise.
        """
        return isinstance(operation, GateOperation)

    def get_measurement_outcome_distribution(
        self, circuit: Circuit, n_samples: Optional[int] = None
    ) -> MeasurementOutcomeDistribution:
        """Get a distribution of measurement outcomes from a given circuit.

        Args:
            circuit: circuit to be sampled
            n_samples: number of times `circuit` should be sampled or None, if
            exact computation should be performed.
        Returns:
            An object representing empirical (for integral n_samples) or theoretical
            (n_samples is None) distribution of measured bitstrings
        """
        if n_samples is None:
            wavefunction = self.get_wavefunction(circuit)
            return create_bitstring_distribution_from_probability_distribution(
                wavefunction.get_probabilities()
            )
        else:
            # Get the expectation values
            measurements = self.run_and_measure(circuit, n_samples)
            return measurements.get_distribution()



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
"""Types commonly encountered in orquestra repositories."""
from numbers import Number
from os import PathLike
from typing import Any, Dict, Protocol, Sequence, Union, runtime_checkable

import numpy as np
import sympy


@runtime_checkable
class Readable(Protocol):
    def read(self, size: int = 0) -> str:
        pass

    def writable(self) -> bool:
        pass


@runtime_checkable
class Writeable(Protocol):
    def write(self, content: str):
        pass

    def writable(self) -> bool:
        pass


AnyPath = Union[str, bytes, PathLike]

LoadSource = Union[Readable, AnyPath]

DumpTarget = Union[Writeable, AnyPath]

Specs = Union[str, Dict]


Parameter = Union[sympy.Symbol, Number]
ParameterizedVector = Union[Sequence[Parameter], np.ndarray]


class SupportsLessThan(Protocol):
    def __lt__(self, other: Any) -> bool:
        """Return result of comparison self < other."""


StateVector = Union[Sequence[complex], np.ndarray]



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
"""Test cases for symbolic_expressions module."""
import pytest
import sympy

from orquestra.quantum.circuits.symbolic.sympy_expressions import (
    SYMPY_DIALECT,
    expression_from_sympy,
)
from orquestra.quantum.circuits.symbolic.translations import translate_expression


@pytest.mark.parametrize(
    "sympy_expression",
    [
        sympy.Symbol("theta"),
        sympy.Symbol("theta") * sympy.Symbol("gamma"),
        sympy.cos(sympy.Symbol("theta")),
        sympy.cos(2 * sympy.Symbol("theta")),
        sympy.exp(sympy.Symbol("x") - sympy.Symbol("y")),
        sympy.cos("phi") + sympy.I * sympy.sin(sympy.Symbol("phi")),
        sympy.Symbol("x") + sympy.Symbol("y") * (2 + 3j),
        sympy.cos(sympy.sin(sympy.Symbol("tau"))),
        sympy.Symbol("x") / sympy.Symbol("y"),
        sympy.tan(sympy.Symbol("theta")),
        2 ** sympy.Symbol("x"),
        sympy.Symbol("y") ** sympy.Symbol("x"),
        sympy.Symbol("x") ** 2,
        sympy.sqrt(sympy.Symbol("x") - sympy.Symbol("y")),
        -5 * sympy.Symbol("x") * sympy.Symbol("y"),
        sympy.Symbol("x") + sympy.Symbol("y") + 9,
    ],
)
def test_translating_tree_from_sympy_to_quil_gives_expected_result(sympy_expression):
    expression = expression_from_sympy(sympy_expression)
    assert translate_expression(expression, SYMPY_DIALECT) - sympy_expression == 0



################################################################################
# © Copyright 2022 Zapata Computing Inc.
################################################################################
import os
import unittest

import numpy as np

from orquestra.quantum.operators import PauliTerm, hermitian_conjugated
from orquestra.quantum.operators._io import (
    convert_dict_to_op,
    convert_op_to_dict,
    get_pauli_strings,
    load_operator,
    load_operator_set,
    save_operator,
    save_operator_set,
)


class Testoperator(unittest.TestCase):
    def setUp(self):
        np.random.seed(0)

    def test_op_to_dict_io(self):
        # Given
        op = PauliTerm("Y0*X1*Z2*X4", 3.0j)
        op += hermitian_conjugated(op)

        # When
        op_dict = convert_op_to_dict(op)
        recreated_op = convert_dict_to_op(op_dict)

        # Then
        self.assertEqual(recreated_op, op)

    def test_operator_io(self):
        # Given
        op = PauliTerm("Y0*X3*Z8*X11", 3.0j)

        # When
        save_operator(op, "op.json")
        loaded_op = load_operator("op.json")

        # Then
        self.assertEqual(op, loaded_op)
        os.remove("op.json")

    def test_operator_set_io(self):
        op1 = PauliTerm("Y0*X3*Z8*X11", 3.0j)
        op2 = PauliTerm("Y0*X1*Z7*X14", 1.0j)

        operator_set = [op1, op2]
        save_operator_set(operator_set, "operator_set.json")
        loaded_operator_set = load_operator_set("operator_set.json")
        for i in range(len(operator_set)):
            self.assertEqual(operator_set[i], loaded_operator_set[i])
        os.remove("operator_set.json")

    def test_op_io(self):
        # Given
        op = PauliTerm("Y0*X1*Z2*X4", 3.0j)

        # When
        save_operator(op, "op.json")
        loaded_op = load_operator("op.json")

        # Then
        self.assertEqual(op, loaded_op)
        os.remove("op.json")

    def test_get_pauli_strings(self):
        operator = (
            PauliTerm("X0*Y1") - 0.5 * PauliTerm("Y1") + 0.5 * PauliTerm.identity()
        )
        constructed_list = get_pauli_strings(operator)
        target_list = ["X0Y1", "Y1", ""]
        self.assertListEqual(constructed_list, target_list)



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
from typing import Tuple

import numpy as np
import numpy.random

from . import _builtin_gates
from ._circuit import Circuit

ONE_QUBIT_NO_PARAMS_GATES = [
    _builtin_gates.X,
    _builtin_gates.Y,
    _builtin_gates.Z,
    _builtin_gates.H,
    _builtin_gates.S,
    _builtin_gates.T,
]

ONE_QUBIT_ONE_PARAM_GATES = [
    _builtin_gates.RX,
    _builtin_gates.RY,
    _builtin_gates.RZ,
    _builtin_gates.PHASE,
]

TWO_QUBITS_NO_PARAMS_GATES = [
    _builtin_gates.CNOT,
    _builtin_gates.CZ,
    _builtin_gates.SWAP,
]

TWO_QUBITS_ONE_PARAM_GATES = [
    _builtin_gates.CPHASE,
]


def create_random_circuit(
    n_qubits: int,
    n_gates: int,
    rng: np.random.Generator,
) -> Circuit:
    """Generates random circuit acting on nqubits with ngates for testing purposes.
    The resulting circuit it saved to file in JSON format under 'circuit.json'.

    Args:
        n_qubits: The number of qubits in the circuit
        n_gates: The number of gates in the circuit
        rng: Numpy random generator

    Returns:
        Generated circuit.
    """
    # Initialize all gates in set, not including RH or ZXZ

    all_gates_lists = np.array(
        [
            ONE_QUBIT_NO_PARAMS_GATES,
            TWO_QUBITS_NO_PARAMS_GATES,
            ONE_QUBIT_ONE_PARAM_GATES,
            TWO_QUBITS_ONE_PARAM_GATES,
        ],
        dtype=object,
    )

    # Loop to add gates to circuit
    circuit = Circuit()
    for _ in range(n_gates):
        # Pick gate type
        gates_list = rng.choice(all_gates_lists)
        gate = rng.choice(gates_list)

        # Pick qubit to act on (control if two qubit gate)
        qubits: Tuple[int, ...]
        if gates_list in [ONE_QUBIT_NO_PARAMS_GATES, ONE_QUBIT_ONE_PARAM_GATES]:
            index = rng.choice(range(n_qubits))
            qubits = (int(index),)
        else:
            indices = rng.choice(range(n_qubits), size=2, replace=False)
            qubits = tuple(int(i) for i in indices)

        if gates_list in [ONE_QUBIT_ONE_PARAM_GATES, TWO_QUBITS_ONE_PARAM_GATES]:
            param = rng.uniform(-np.pi, np.pi, size=1)
            gate = gate(float(param))

        circuit += gate(*qubits)
    return circuit



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
"""Utilities for converting symbolic expressions between different dialects."""

from functools import reduce
from numbers import Number
from typing import Any, Callable, Dict, Iterable, NamedTuple


class Symbol(NamedTuple):
    """Abstract symbol."""

    name: str


class FunctionCall(NamedTuple):
    """Represents abstract function call."""

    name: str
    args: Iterable["Expression"]


# Note that mypy does not support recursive types, so for now Expression is set
# to Any. See mypy #731 for details.
Expression = Any
# Expression = Union[Symbol, FunctionCall, Number]


class ExpressionDialect(NamedTuple):
    """Dialect of arithmetic expression.

    This is to group information on how to transform expression given in
    our native representation into some representation in external
    library (e.g. Sympy).
    """

    symbol_factory: Callable[[Symbol], Any]
    number_factory: Callable[[Number], Any]
    known_functions: Dict[str, Callable[..., Any]]


def reduction(operator):
    def _reduction(*args):
        return reduce(operator, args)

    return _reduction



################################################################################
# © Copyright 2021-2022 Zapata Computing Inc.
################################################################################
import pytest
import sympy

from orquestra.quantum.circuits import natural_key, natural_key_revlex
from orquestra.quantum.circuits.symbolic import natural_key_fixed_names_order


@pytest.mark.parametrize(
    "unordered_symbols, expected_ordered_symbols",
    [
        (
            sympy.symbols("theta_10, theta_2, theta_1"),
            sympy.symbols("theta_1, theta_2, theta_10"),
        ),
        (
            sympy.symbols("beta_10, theta_1, beta_2, theta_2"),
            sympy.symbols("beta_2, beta_10, theta_1, theta_2"),
        ),
        (sympy.symbols("gamma, beta, alpha"), sympy.symbols("alpha, beta, gamma")),
    ],
)
def test_natural_key_orders_symbols_as_expected(
    unordered_symbols, expected_ordered_symbols
):
    assert sorted(unordered_symbols, key=natural_key) == list(expected_ordered_symbols)


@pytest.mark.parametrize(
    "unordered_symbols, expected_ordered_symbols",
    [
        (
            sympy.symbols("theta_10, theta_2, theta_1"),
            sympy.symbols("theta_1, theta_2, theta_10"),
        ),
        (
            sympy.symbols("beta_10, theta_1, beta_2, theta_2"),
            sympy.symbols("theta_1, beta_2, theta_2, beta_10"),
        ),
        (sympy.symbols("gamma, beta, alpha"), sympy.symbols("alpha, beta, gamma")),
    ],
)
def test_natural_key_revlex_orders_symbols_as_expected(
    unordered_symbols, expected_ordered_symbols
):
    assert sorted(unordered_symbols, key=natural_key_revlex) == list(
        expected_ordered_symbols
    )


@pytest.mark.parametrize(
    "names_order, unordered_symbols, expected_ordered_symbols",
    [
        (
            ("gamma", "beta"),
            sympy.symbols("beta_0, gamma_0, beta_3, gamma_3, beta_4"),
            sympy.symbols("gamma_0, beta_0, gamma_3, beta_3, beta_4"),
        ),
        (
            ("theta", "beta", "gamma"),
            sympy.symbols("gamma_0, beta_0, theta_0, beta_1, theta_1, gamma_1"),
            sympy.symbols("theta_0, beta_0, gamma_0, theta_1, beta_1, gamma_1"),
        ),
    ],
)
def test_natural_key_fixed_names_order_orders_symbols_ax_expected(
    names_order, unordered_symbols, expected_ordered_symbols
):
    key = natural_key_fixed_names_order(names_order)
    assert sorted(unordered_symbols, key=key) == list(expected_ordered_symbols)
